;Chong Carl
;SPARC SERVO.ASM

;This routine drives four standard RC servos connected to port C (easily changed)
;The number of servos could be reduced fairly easily by simply changing the test
;in the TIM1_OVF of ISR.ASM

;The servo control bytes are stored as SERVO_X(1-8) in SRAM, Modify as you like,
;the servo position will be updated when it's turn comes up next.

;Servo outputs are activated sequentially, so that no more than one servo is active
;at any time. The servos may take some time to physically slew to their new positions,
;but we are only talking to one of them at a time. This is also how standard R/C 
;equipment works, so there's nothing new here.

;PORTC's current status tells us whether a servo is currently active, there should only
;ever be a single high bit on this port.

;Servo_Control determines wether any servos should be active. Every time timer1 expires,
;it shifts the Servo_Control byte. If the active bit shifts into carry, then it re-inits
;Servo_Control. Setting Servo_Control to zero will cause the system to quit outputting 
;servo pulses.

;Frame_Delay is a timer byte, decremented by the Timer0 int every millisecond. It is 
;initialized by the timer1 ISR to some value (set in EQUATES.INC). Before we output the
;first servo pulse, we must see Frame_Delay decrement to zero. This assures a constant
;frame rate, no matter what the widths of the servo pulses is, since the timers run 
;concurrently.

;*********************************************************************
;Hardware assignments: Any port, and all of a port, as written.
.equ	Servo_DATA=PORTC;Where the bits show up
.equ	Servo_DDR=DDRC	;So we can set the bits to output
;*********************************************************************
Init_Servo:
	rcall	Servo_Set		;Set initial widths for the servos (SERVO.ASM)
					;When there's a real application running, this
					;might be redundant.
	ldi	TEMP,$01		;
	sts	Servo_Control,TEMP	;Set servo 0(1) active

	ldi	TEMP,FRAME_RATE		;
	sts	FRAME_DELAY,TEMP	;Set the frame rate delay

	ldi	TEMP,$00		;
	out	Servo_Data,TEMP		;Set servo outputs inactive
	ret
;*********************************************************************
Servo_Lights_On:
	;Lights detected so run until it it dark...
	;The Servo_Control byte tells us which servo output will be active.
	;If it's zero, then no servos are active, and we don't send any output.
	;Otherwise, it's just a matter of picking which one to send a signal to.
	;Without any pulses, the servos go to idle and may be pushed by external forces.

	;First, we check if there are any servo outs currently running, if so,
	;we want to let them expire (TIM1_OVF in ISR.ASM) before doing anything else.
	;This means that the vast majority of the time, we just do this test
	;and bail, preserving CPU time for other things! :)

	in	TEMP,Servo_Data		;Get current servo outputs
	and	TEMP,TEMP		;Anything busy?
	brne	Servo_Quit		;Yes, just leave them alone!
	
	;Now, we check wether ANY servos SHOULD be active. There's only one
	;bit on in here at any time, starting with bit zero. (Servo_1)
	;If no bits are active, then we can just bail, since we aren't 
	;outputting any servos at this time (servos are free-floating then,
	;and can be moved since they aren't seeking any position actively)

	lds	TEMP,Servo_Control	;Get from SRAM
	and	TEMP,TEMP		;Is it zero?
	breq	Servo_Quit		;Nope, just pass

	;We've determined that a servo should be active
	;If it's servo 1, then we want to check if the frame delay has expired.
	;FRAME_DELAY is decremented to zero by timer 0 (the 1ms opsys tick)
	;If it hasn't expired, then again we bail out. If it has then we go ahead
	;and start servo 1, and reload FRAME_DELAY. Note that this test is only done
	;before SERVO_1 is output, it's not a delay between all channels, just between 
	;SERVO_8 and SERVO_1.

	cpi	TEMP,$01		;Is it servo 1
	brne	Servo_Pulse		;If not, just continue
	
	lds	TEMP,Frame_Delay	;Get the frame delay byte
	and	TEMP,TEMP		;If so, check if the servo frame delay has expired
	brne	Servo_Quit		;If not, then don't do anything till it does.

	;If this is servo 1, and FRAME_DELAY has expired, then reload FRAME_DELAY, and
	;go ahead with the servo pulses
	ldi	TEMP,FRAME_RATE		;Dec'd by Timer 0
	sts	Frame_Delay,TEMP	;Put it back

	;Output on the servo channels.
	rcall	Servo_Run
	rjmp	Servo_Pulse		;Go output a pulse
	ret

Servo_Lights_Off:
	;Lights NOT detected so walk around...
	;The Servo_Control byte tells us which servo output will be active.
	;If it's zero, then no servos are active, and we don't send any output.
	;Otherwise, it's just a matter of picking which one to send a signal to.
	;Without any pulses, the servos go to idle and may be pushed by external forces.

	;First, we check if there are any servo outs currently running, if so,
	;we want to let them expire (TIM1_OVF in ISR.ASM) before doing anything else.
	;This means that the vast majority of the time, we just do this test
	;and bail, preserving CPU time for other things! :)

	in	TEMP,Servo_Data		;Get current servo outputs
	and	TEMP,TEMP		;Anything busy?
	brne	Servo_Quit		;Yes, just leave them alone!
	
	;Now, we check wether ANY servos SHOULD be active. There's only one
	;bit on in here at any time, starting with bit zero. (Servo_1)
	;If no bits are active, then we can just bail, since we aren't 
	;outputting any servos at this time (servos are free-floating then,
	;and can be moved since they aren't seeking any position actively)

	lds	TEMP,Servo_Control	;Get from SRAM
	and	TEMP,TEMP		;Is it zero?
	breq	Servo_Quit		;Nope, just pass

	;We've determined that a servo should be active
	;If it's servo 1, then we want to check if the frame delay has expired.
	;FRAME_DELAY is decremented to zero by timer 0 (the 1ms opsys tick)
	;If it hasn't expired, then again we bail out. If it has then we go ahead
	;and start servo 1, and reload FRAME_DELAY. Note that this test is only done
	;before SERVO_1 is output, it's not a delay between all channels, just between 
	;SERVO_8 and SERVO_1.

	cpi	TEMP,$01		;Is it servo 1
	brne	Servo_Pulse		;If not, just continue
	
	lds	TEMP,Frame_Delay	;Get the frame delay byte
	and	TEMP,TEMP		;If so, check if the servo frame delay has expired
	brne	Servo_Quit		;If not, then don't do anything till it does.

	;If this is servo 1, and FRAME_DELAY has expired, then reload FRAME_DELAY, and
	;go ahead with the servo pulses
	ldi	TEMP,FRAME_RATE		;Dec'd by Timer 0
	sts	Frame_Delay,TEMP	;Put it back

	;Output on the servo channels.
	rcall	Servo_Walk
	rjmp	Servo_Pulse		;Go output a pulse

Servo_Quit:
	ret	

Servo_Pulse:
	;Figure out which servo time to pick up, 0-7 (in temp)
	ldi	TEMP2,0			;Init TEMP2
	lds	TEMP,Servo_Control	;

	;As we shift our copy of the Servo_Control right toward carry, Temp2 is 
	;incremented to form an offset, which we will add to the base of the 
	;servo time widths in RAM to obtain the right width for this servo

Servo_PTR:				;
	lsr	TEMP			;Shift the servo bit twoard carry
	brcs	Servo_Point		;If carry, then we have our servo	
	inc	TEMP2			;Otherwise inc, and 
	rjmp	Servo_PTR		;try again

	;Now we have the index, set the main pointer in R30,31, and grab the 
	;proper servo width byte from SRAM
	;TEMP2 has an integer offset

Servo_Point:
	;Point at the base of the servo times in RAM
	ldi	ZL,low(SERVO_1)		;Point at the lowest servo
	ldi	ZH,high(SERVO_1)	;

	;Add the offset to pick SERVO_X
	clc				;			
	adc	ZL,TEMP2		;Add the servo number (0-7)

	;Handle carry if the location crosses a boundary
	brcc	Servo_Point_B		;No carry, we're done
	inc	ZH			;else inc the high byte of the pointer
	
	;Retrieve the servo width data
Servo_Point_B:
	ld	TEMP3,Z			;Get the servo time

	;At this point, TEMP2 has the servo number (0-7)
	;Temp3 has the proportional width that we would like to set it to.
	;Get the base pulsewidth out of the min table

	ldi	ZL,low(Servo_Min_Table*2)  ;Point at the minimum widths table
	ldi	ZH,high(Servo_Min_Table*2) ;
	clc				;
	mov	TEMP,TEMP2		;Take the servo index
	lsl	TEMP			;Mult by 2
	adc	ZL,TEMP			;Add to Z low
	brcc	Servo_Point_C		;No carry, we're done
	inc	ZH			;else inc the high byte of the pointer

	;Retrieve the servo base width data

Servo_Point_C:
	lpm			;Table to R0
	mov	TEMP5,R0	;Put it in TEMP5 (low)
	ld	TEMP4,Z+	;Move the pointer
	lpm			;Table to R0 (other half)
	mov	TEMP4,R0	;Put it in TEMP4 (high)

Servo_Point_D:

	;At this point, TEMP4,5 has FFFF - the servo base value for this servo.
	;Now we have to apply the width value to servo_range, to figure out 
	;where to point it.
	;We divide this range value by 256, then multiply by the servo value.(0-255)

	;Get the servo_range for this servo, stored as a 16 bit value into TEMP6(h),TEMP7(l)
	ldi	ZL,low(Servo_Range_Table*2)	;Point at the lowest servo
	ldi	ZH,high(Servo_Range_Table*2)	;

	clc				;
	mov	TEMP,TEMP2		;
	lsl	TEMP			;			
	adc	ZL,TEMP			;Add the servo number (0-7)
	brcc	Servo_Point_E		;No carry, we're done
	inc	ZH			;else inc the high byte of the pointer

Servo_Point_E:
	lpm				;Table to R0
	mov	TEMP7,R0		;Get the servo base width in microseconds
	ld	TEMP6,Z+		;Move the pointer
	lpm				;Table to R0
	mov	TEMP6,R0		;Low
	
	;Divide the servo_range by 256, result in ???
	ldi	TEMP,8			;
	mov	LOOP,TEMP		;

Servo_Point_EA:
	lsr	TEMP6			;/2
	ror	TEMP7			;
	dec	LOOP			;8 times
	brne	Servo_Point_EA		;
	;129 is as big as it gets, so it's an 8 bit quantity now, in TEMP7

	;Multiply by the servo width value, result in TEMP4,5
	mov	LOOP,TEMP3		;	
	;Subtract the 16 bit result from the timer

Servo_Point_EB:
	clc				;
	sub	TEMP5,TEMP7		;
	brcc	Servo_Point_EC		;
	dec	TEMP4			;

Servo_Point_EC:
	dec	LOOP			;
	brne	Servo_Point_EB
	;Now TEMP4,5 contains a 16 bit reload value representing the base
	;width, plus the proportional part

	out	TCNT1H,TEMP4		;Timer high byte
	out	TCNT1L,TEMP5		;Timer low byte

	;Start timer 1
	in	TEMP,TCCR1B		;Get the timer control register
	andi	TEMP,$F8		;Mask out the timer bits
	ori	TEMP,$01		;
	out	TCCR1B,TEMP		;Make it so.
	
	;Start the servo output active
	lds	TEMP,Servo_Control	;Get the active servo bit
	out	Servo_Data,TEMP		;Start the output
	ldi	TEMP,$FF		;Make them all outputs
	out	Servo_DDR,TEMP		

	ret				

Servo_Set:
	;Reset the servos to neutral
	ldi	TEMP,$FF	
	sts	Servo_1,TEMP	;Set a full wide pulse

	ldi	TEMP,$FF
	sts	Servo_2,TEMP	;Set a full wide pulse

	ldi	TEMP,$FF	
	sts	Servo_3,TEMP	;Set a full wide pulse

	ldi	TEMP,$FF	
	sts	Servo_4,TEMP	;Set a full wide pulse
	
	ret

Servo_Run:
	;Cricket is running

	lds	TEMP,Servo_1		;Change this to SERVO_X as desired
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	sts	SERVO_1,TEMP		;

	lds	TEMP,Servo_2		;Change this to SERVO_X as desired
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	sts	SERVO_2,TEMP		;

	lds	TEMP,Servo_3		;Change this to SERVO_X as desired
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	sts	SERVO_3,TEMP		;

	lds	TEMP,Servo_4		;Change this to SERVO_X as desired
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	inc	TEMP			;
	sts	SERVO_4,TEMP		;

	ret				;

Servo_Walk:
	;Cricket is walking

	lds	TEMP,Servo_1		;Change this to SERVO_X as desired
	inc	TEMP			;
	inc	TEMP			;
	sts	SERVO_1,TEMP		;

	lds	TEMP,Servo_2		;Change this to SERVO_X as desired
	inc	TEMP			;
	inc	TEMP			;
	sts	SERVO_2,TEMP		;

	lds	TEMP,Servo_3		;Change this to SERVO_X as desired
	inc	TEMP			;
	inc	TEMP			;
	sts	SERVO_3,TEMP		;

	lds	TEMP,Servo_4		;Change this to SERVO_X as desired
	inc	TEMP			;
	inc	TEMP			;
	sts	SERVO_4,TEMP		;

	ret				;

